/* MERG Content-to-template merger
   Command line:
      MERG [switches] template content1 content2 ...

   Valid switches:
      h        allow/Suppress HTML tags in content
      n        Keep/discard name
      l        long file name Acceptable/truncate
      e ERROR  Output error messages to ERROR
               instead of STD_ERR
      q        Quiet mode


mparset.c Template parser and tokenizer
This file contains the template parser main function call
(ParseTemplate), as well as internal "helper" functions.
The parser is divided into two passes. The first pass
breaks the template character stream into pseudo-tokens.
The second pass moves through the token stream, and
extracts MERG-specific commands, and takes note of them.
A function (EvalTag) is used by the first pass parser to
determine an HTML markup tag's relevance. If the tag is
a MERG command tag, the token is marked as such.
*/

#include "merg.h"

/* Module-wide string for error reporting */
static char errstr[256];

int EvalTag (char * tagcontent, int prev)
{
	int offset;

	if(tagcontent[2] == '-' && tagcontent[3] == '-')
	{
		for(offset=4; isspace(tagcontent[offset]); offset++)
			;
		
		/* This is a comment, or a MERG command */
		if(toupper(tagcontent[offset])=='M'
		&& toupper(tagcontent[offset+1])=='E'
		&& toupper(tagcontent[offset+2])=='R'
		&& toupper(tagcontent[offset+3])=='G'){
			return TKN_T_COMMAND;
		}
		else
			return TKN_COMMENT;
	}
	else
		return TKN_HTML;

	/* Fallback; this won't happen */

	return prev;
}

int CheckScript(char * tagtxt)
{
	if(tolower(tagtxt[1])=='s'
	&& tolower(tagtxt[2])=='c'
	&& tolower(tagtxt[3])=='r'
	&& tolower(tagtxt[4])=='i'
	&& tolower(tagtxt[5])=='p'
	&& tolower(tagtxt[6])=='t')
		return -1;
	else
		return 0;
}

int CheckScriptEnd(char * tagtxt)
{
	if((tagtxt[1])=='/'
	&& tolower(tagtxt[2])=='s'
	&& tolower(tagtxt[3])=='c'
	&& tolower(tagtxt[4])=='r'
	&& tolower(tagtxt[5])=='i'
	&& tolower(tagtxt[6])=='p'
	&& tolower(tagtxt[7])=='t')
		return -1;
	else
		return 0;
}

TOKENSTREAM * ParseTemplatePass1(char *filename, int flags)
{
	FILE * src;
	TOKENSTREAM * resultstream;
	char c;
	char buff[BUFFSIZE];
	TOKEN current;
	int i, prev, b = 0;

	int dirty=0, insec=0, inscript=0, inscrtag=0;
	int nlcount = 0;
	
	src = fopen(filename, "r");
	if (!src){
		if(flags & B_DEBUG){
			sprintf(errstr, "ParseTemplate[%i]:", __LINE__);
			ErrOut(errstr);
		}
		sprintf(errstr, "Error opening file %s\n", filename);
		ErrOut(errstr);
		return NULL;
	}
	
	resultstream = CreateStream();
	if(resultstream == NULL){
		if (flags & B_DEBUG){
			sprintf(errstr, "ParseTemplate[%i]:", __LINE__);
			ErrOut(errstr);
		}
		sprintf(errstr, "Error creating token stream\n");
		ErrOut(errstr);
		return NULL;
	}
	
	for(i=0; i < BUFFSIZE; i++)
		buff[i] = 0;

	CreateToken(&current);
	TokenType(&current, TKN_TEXT);

	while((c=fgetc(src)) != EOF)
	{
		if (c=='\n' && !insec && !inscript){
			if(dirty){
				nlcount++;
				TokenContent(&current, buff);
				
				for (i = 0; i < BUFFSIZE; i++)
					buff[i] = 0;
				b = 0;
				AddToken (resultstream, &current);
				prev = TokenType(&current, 0);
				TokenType(&current, TKN_NEWLINE);
				TokenContent(&current, "\n");
				AddToken (resultstream, &current);
				TokenType(&current, prev);
			}
			else if(!dirty){
				prev = TokenType(&current, 0);
				TokenType(&current, TKN_NEWLINE);
				TokenContent(&current, "\n");
				AddToken (resultstream, &current);
				TokenType(&current, prev);
			}

			continue;
		}
		
		/* Discard CRs */
		if (c=='\r')
			continue;

		if (c=='<'){
			if(dirty && !inscript){
				TokenContent(&current, buff);
				for (i = 0; i < BUFFSIZE; i++)
					buff[i] = 0;
				b = 0;
				AddToken (resultstream, &current);
				TokenType(&current, TKN_MARKUP);
				dirty=0;
				insec=1;
			}
			else if(!dirty && !inscript){
				TokenType(&current, TKN_MARKUP);
				insec=1;
			}
			else if(dirty && inscript){
				inscrtag=strlen(buff);
			}
		}

		buff[b] = c;
		b++;
		if(b > BUFFSIZE){
			ErrOut("FATAL ERROR: Buffer overflow, dumping buffer contents...\n");
			sprintf(errstr, "%s\n", buff);
			ErrOut(errstr);
			ErrOut("Terminating MERG... (255)\n");
			exit(255);
		}
		dirty=1;

		if (c=='>'){
			if (dirty && insec && !inscript){
				if (CheckScript(buff)){
					inscript=1;
				}
				TokenContent(&current, buff);
				TokenType(&current, EvalTag (TokenContent(&current, NULL), TokenType(&current, 0)));
				for (i = 0; i < BUFFSIZE; i++)
					buff[i] = 0;
				b = 0;
				AddToken (resultstream, &current);
				TokenType(&current, TKN_TEXT);
				insec=0;
			}
			else if (dirty && inscript){
				if(CheckScriptEnd(&buff[inscrtag])){
					inscript=0;
				}
				else
					inscrtag=0;
			}
		}
	}

	if (flags & B_DEBUG)
		printf("%i tokens created, %i NEWLINEs.", resultstream->count, nlcount);

	fclose(src);

	return resultstream;
}

TOKENSTREAM * ParseTemplatePass2(TOKENSTREAM * parse1stream)
{
	TOKEN * current;
	char * mergcmd, * name;
	int i = 0, j = 0, offset;

	current = FindTokenType(parse1stream, TKN_T_COMMAND);
	if(current == NULL)
		return NULL;


	do {
		i = 0;
		j = 0;
		mergcmd = (char *) MemAlloc (sizeof(char) * (strlen(current->szContent)+1));
		mergcmd = strcpy(mergcmd, current->szContent);

		for(j = 4; isspace(mergcmd[j]); j++)
			;
		offset = j + 5;
		j += 5;
		while(mergcmd[j] != '\0' && !isspace(mergcmd[j]) && mergcmd[j] != '-'){
			i++;
			j++;
		}

		name = (char *) MemAlloc (sizeof(char) * i);
		for(j=0; j < i; j++)
			name[j] = tolower((mergcmd + offset)[j]);
		name[i] = 0;

		TokenName (current, name);

		AttachName(parse1stream, OffsetToken(parse1stream));

		MemFree(name);
		MemFree(mergcmd);
	} while(current = FindTokenType(parse1stream, 0));

	return parse1stream;
}

TOKENSTREAM * ParseTemplate(char *filename, int flags)
{
	TOKENSTREAM * rstream;
	
	if (flags & B_DEBUG)
		ErrOut("Pass 1:");
	rstream = ParseTemplatePass1(filename, flags);
	
	if(rstream == NULL){
		if (flags & B_DEBUG)
			ErrOut("Error!\n");
		sprintf(errstr, "Template %s: File not found or error opening file.\n", filename);
		ErrOut(errstr);
		return NULL;
	}
	else	if (flags & B_DEBUG){
		ErrOut("Okay.\n");
	}

	if (flags & B_DEBUG)
		ErrOut("Pass 2:");
	rstream = ParseTemplatePass2(rstream);

	
	if(rstream == NULL){
		if (flags & B_DEBUG)
			ErrOut("Error!\n");
		sprintf(errstr, "Template %s: No valid MERG statements found.\n", filename);
		ErrOut(errstr);
		return NULL;
	}
	else if (flags & B_DEBUG)
		ErrOut("Okay.\n");
	
	if (flags & B_DEBUG)
		ErrOut("Template parsed okay, returning...\n\n");

	SetStreamName(rstream, filename);

	return rstream;
}
